@@ -13,9 +13,9 @@ pub struct World {
     pub(crate) entities: Vec<EntityMetadata>,
     free_list: Vec<u32>,
     pub(crate) archetypes: Vec<Archetype>,
-    insert_map: IndexTypeIdMap,
-    remove_map: IndexTypeIdMap,
-    transfer_map: U32PairMap,
+    insert_map: IndexTypeIdMap<u16>,
+    remove_map: IndexTypeIdMap<u16>,
+    transfer_map: IndexTagMap<u16>,
 }
 
 impl Default for World {
@@ -145,9 +145,9 @@ impl World {
 }
 
 impl World {
-    pub(crate) fn tag_gen(&self) -> (u32, u32) {
+    pub(crate) fn tag_gen(&self) -> (u32, u16) {
         debug_assert!(!self.archetypes.is_empty());
-        (self.tag, self.archetypes.len() as u32)
+        (self.tag, self.archetypes.len() as u16)
     }
 
     /// Insert components for a given [Entity].
@@ -176,10 +176,12 @@ impl World {
         let new_ty = if let Some(ty) = self.insert_map.get(&(meta.ty, TypeId::of::<B>())) {
             *ty
         } else {
-            Self::insert_cold::<B>(
+            Self::insert_cold(
                 &mut self.archetypes,
                 &mut self.insert_map,
                 &mut self.remove_map,
+                TypeId::of::<B>(),
+                B::insert,
                 meta.ty,
             )
         };
@@ -196,22 +198,21 @@ impl World {
 
     #[cold]
     #[inline(never)]
-    fn insert_cold<B>(
+    fn insert_cold(
         archetypes: &mut Vec<Archetype>,
-        insert_map: &mut IndexTypeIdMap,
-        remove_map: &mut IndexTypeIdMap,
-        old_ty: u32,
-    ) -> u32
-    where
-        B: Bundle,
-    {
+        insert_map: &mut IndexTypeIdMap<u16>,
+        remove_map: &mut IndexTypeIdMap<u16>,
+        type_id: TypeId,
+        insert: fn(&mut TypeMetadataSet),
+        old_ty: u16,
+    ) -> u16 {
         let mut types = archetypes[old_ty as usize].types();
-        B::insert(&mut types);
+        insert(&mut types);
 
         let new_ty = Self::get_or_insert(archetypes, types);
 
-        insert_map.insert((old_ty, TypeId::of::<B>()), new_ty);
-        remove_map.insert((new_ty, TypeId::of::<B>()), old_ty);
+        insert_map.insert((old_ty, type_id), new_ty);
+        remove_map.insert((new_ty, type_id), old_ty);
 
         new_ty
     }
@@ -240,10 +241,12 @@ impl World {
         let new_ty = if let Some(ty) = self.remove_map.get(&(meta.ty, TypeId::of::<B>())) {
             *ty
         } else {
-            Self::remove_cold::<B>(
+            Self::remove_cold(
                 &mut self.archetypes,
                 &mut self.insert_map,
                 &mut self.remove_map,
+                TypeId::of::<B>(),
+                B::remove,
                 meta.ty,
             )?
         };
@@ -262,44 +265,43 @@ impl World {
 
     #[cold]
     #[inline(never)]
-    fn remove_cold<B>(
+    fn remove_cold(
         archetypes: &mut Vec<Archetype>,
-        insert_map: &mut IndexTypeIdMap,
-        remove_map: &mut IndexTypeIdMap,
-        old_ty: u32,
-    ) -> Option<u32>
-    where
-        B: Bundle,
-    {
+        insert_map: &mut IndexTypeIdMap<u16>,
+        remove_map: &mut IndexTypeIdMap<u16>,
+        type_id: TypeId,
+        remove: fn(&mut TypeMetadataSet) -> Option<()>,
+        old_ty: u16,
+    ) -> Option<u16> {
         let mut types = archetypes[old_ty as usize].types();
-        B::remove(&mut types)?;
+        remove(&mut types)?;
 
         let new_ty = Self::get_or_insert(archetypes, types);
 
-        remove_map.insert((old_ty, TypeId::of::<B>()), new_ty);
-        insert_map.insert((new_ty, TypeId::of::<B>()), old_ty);
+        remove_map.insert((old_ty, type_id), new_ty);
+        insert_map.insert((new_ty, type_id), old_ty);
 
         Some(new_ty)
     }
 
-    fn get_or_insert(archetypes: &mut Vec<Archetype>, types: TypeMetadataSet) -> u32 {
+    fn get_or_insert(archetypes: &mut Vec<Archetype>, types: TypeMetadataSet) -> u16 {
         let pos = archetypes
             .iter_mut()
             .position(|archetype| archetype.match_(&types));
 
         if let Some(pos) = pos {
-            pos as u32
+            pos as u16
         } else {
             let len = archetypes.len();
-            assert!(len < u32::MAX as usize);
+            assert!(len < u16::MAX as usize);
 
             archetypes.push(Archetype::new(types));
 
-            len as u32
+            len as u16
         }
     }
 
-    unsafe fn move_(&mut self, id: u32, old_ty: u32, new_ty: u32, old_idx: u32) -> u32 {
+    unsafe fn move_(&mut self, id: u32, old_ty: u16, new_ty: u16, old_idx: u32) -> u32 {
         let old_archetype = &mut *self.archetypes.as_mut_ptr().add(old_ty as usize);
         let new_archetype = &mut *self.archetypes.as_mut_ptr().add(new_ty as usize);
 
@@ -396,12 +398,12 @@ impl World {
     fn transfer_cold(
         archetypes: &[Archetype],
         other_archetypes: &mut Vec<Archetype>,
-        transfer_map: &mut U32PairMap,
-        other_transfer_map: &mut U32PairMap,
+        transfer_map: &mut IndexTagMap<u16>,
+        other_transfer_map: &mut IndexTagMap<u16>,
         tag: u32,
         other_tag: u32,
-        old_ty: u32,
-    ) -> u32 {
+        old_ty: u16,
+    ) -> u16 {
         let types = archetypes[old_ty as usize].types();
 
         let new_ty = Self::get_or_insert(other_archetypes, types);
@@ -500,9 +502,11 @@ impl World {
     where
         C: 'static,
     {
-        if TypeId::of::<C>() == TypeId::of::<Entity>() {
-            panic!("Entity cannot be accessed mutably");
-        }
+        assert_ne!(
+            TypeId::of::<C>(),
+            TypeId::of::<Entity>(),
+            "Entity cannot be accessed mutably"
+        );
 
         let meta = &self.entities[ent.id as usize];
         assert_eq!(ent.gen, meta.gen, "Entity is stale");
@@ -521,7 +525,7 @@ pub struct Entity {
 #[derive(Clone, Copy)]
 pub struct EntityMetadata {
     pub gen: NonZeroU32,
-    pub ty: u32,
+    pub ty: u16,
     pub idx: u32,
 }
 
@@ -568,9 +572,11 @@ macro_rules! impl_bundle_for_tuples {
 
             fn remove(types: &mut TypeMetadataSet) -> Option<()> {
                 $(
-                    if TypeId::of::<$types>() == TypeId::of::<Entity>() {
-                        panic!("Entity cannot be removed");
-                    }
+                    assert_ne!(
+                        TypeId::of::<$types>(),
+                        TypeId::of::<Entity>(),
+                        "Entity cannot be removed"
+                    );
 
                     types.remove::<$types>()?;
                 )+
@@ -595,13 +601,13 @@ macro_rules! impl_bundle_for_tuples {
 
 impl_bundle_for_tuples!(A, B, C, D, E, F, G, H, I, J);
 
-type IndexTypeIdMap = HashMap<(u32, TypeId), u32, BuildHasherDefault<IndexTypeIdHasher>>;
+type IndexTypeIdMap<V> = HashMap<(u16, TypeId), V, BuildHasherDefault<IndexTypeIdHasher>>;
 
 #[derive(Default)]
 struct IndexTypeIdHasher(u64);
 
 impl Hasher for IndexTypeIdHasher {
-    fn write_u32(&mut self, val: u32) {
+    fn write_u16(&mut self, val: u16) {
         self.0 = 1 + val as u64;
     }
 
@@ -618,12 +624,16 @@ impl Hasher for IndexTypeIdHasher {
     }
 }
 
-type U32PairMap = HashMap<(u32, u32), u32, BuildHasherDefault<U32PairHasher>>;
+type IndexTagMap<V> = HashMap<(u16, u32), V, BuildHasherDefault<IndexTagHasher>>;
 
 #[derive(Default)]
-struct U32PairHasher(u64);
+struct IndexTagHasher(u64);
+
+impl Hasher for IndexTagHasher {
+    fn write_u16(&mut self, val: u16) {
+        self.0 = val as u64;
+    }
 
-impl Hasher for U32PairHasher {
     fn write_u32(&mut self, val: u32) {
         self.0 = self.0 << 32 | val as u64;
     }
@@ -941,12 +951,12 @@ mod tests {
     }
 
     #[test]
-    fn u32_pair_hasher_yields_uniformly_distributed_lower_bits() {
+    fn index_tag_hasher_yields_uniformly_distributed_lower_bits() {
         let mut histogram = [0; 128];
 
         for i in 0..1024 {
             for j in 0..128 {
-                let mut hasher = U32PairHasher::default();
+                let mut hasher = IndexTagHasher::default();
                 hasher.write_u32(i);
                 hasher.write_u32(j);
                 let hash = hasher.finish();
